/*
 ****************************************************************************
 * Ldap Synchronization Connector provides tools to synchronize
 * electronic identities from a list of data sources including
 * any database with a JDBC connector, another LDAP directory,
 * flat files...
 *
 *                  ==LICENSE NOTICE==
 * 
 * Copyright (c) 2008, LSC Project 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:

 *    * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * Neither the name of the LSC Project nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *                  ==LICENSE NOTICE==
 *
 *               (c) 2008 - 2009 LSC Project
 *         Sebastien Bahloul <seb@lsc-project.org>
 *         Thomas Chemineau <thomas@lsc-project.org>
 *         Jonathan Clarke <jon@lsc-project.org>
 *         Remy-Christophe Schermesser <rcs@lsc-project.org>
 ****************************************************************************
 */
package org.lsc.service;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.lsc.objects.organizationalUnit;
import org.lsc.objects.flat.fOrganizationalUnit;
import org.lsc.persistence.DaoConfig;

import com.ibatis.sqlmap.client.SqlMapClient;

/**
 * @author Sebastien Bahloul <seb@lsc-project.org>
 */
public abstract class OrganizationalUnitJdbcService implements ISrcService {

	private Logger LOGGER = Logger.getLogger(OrganizationalUnitJdbcService.class);

	protected SqlMapClient sqlMapper;
	
    protected Vector<String> orderedOu;
    protected Map<String,List<String>> ouChildren;
    protected Map<String,fOrganizationalUnit> ou;
    protected boolean init ;
    protected int lastOuMemberNb ;

    /** Following values must be filled while instanciated */
	protected String sqlFunctionNameForGetMembersList;
	protected String sqlFunctionNameForGetObject;
	protected String sqlFunctionNameForGetObjectList;
	protected String sqlFunctionNameForGetParentObjectId;
	protected String baseDn;

// Constructors

	public OrganizationalUnitJdbcService() {
        this.orderedOu = new Vector<String>();
        this.ouChildren = new HashMap<String,List<String>>();
        this.ou = new HashMap<String,fOrganizationalUnit>();
        this.init = false ;
        this.lastOuMemberNb = -1 ;
		sqlMapper = DaoConfig.getSqlMapClient();
	}

// Accessors

    /**
     * Get each child of a specified parent id structure
     * @param parentId
     * @return
     */
    public List<fOrganizationalUnit> getChildren(String ouId) {
    	if (init) {
	        Iterator childrenIdIter = ((List) ou.get(ouId)).iterator();
	        List<fOrganizationalUnit> children = new ArrayList<fOrganizationalUnit>();
	        while(childrenIdIter.hasNext()) {
	            String childrenId = (String) childrenIdIter.next();
	            children.add(ou.get(childrenId));
	        }
	        return children;
    	}
    	return new ArrayList<fOrganizationalUnit>();
    }

    /**
     * Get list of a all ou's id.
     * @return One list of all ou's id found in database
     */
	public final List getList() {
		if (init) {
			return orderedOu.subList(0, orderedOu.size());
		}
		return new Vector();
    }

    /**
     * Get the parent id of a specified organizationalUnit.
     * @param ouId
     * @return An organizationalUnit object.
     */
    public organizationalUnit getParent(String ouId) {
        List parent = _getParent(ouId);
        if (parent.size() <= 0) {
            LOGGER.warn("There is a big problem with the structure identified by \"" + ouId
                    + "\": no parent structure !");
        } else if (parent.size() >= 1) {
            LOGGER.warn("There is a big problem with the structure identified by \"" + ouId
                    + "\": more than one parent !");
        } else {
            return (organizationalUnit) parent.get(0);
        }
        return null;
    }

// Public methods

	/**
	 * Initialize tree.
	 * Do SQL request on Database.
	 */
    public void initialize() {
    	init = true;
        List<String> topLevelOuIds = new ArrayList<String>();
        List originalOuIdList = _getOuIdList();

        Iterator originalOuIdListIter = originalOuIdList.iterator();
        while (originalOuIdListIter.hasNext()) {
            String originalOuId = (String) originalOuIdListIter.next();
            List parent = _getParent(originalOuId);
            while (parent.contains(originalOuId)) {
                parent.remove(originalOuId);
            }

            if (parent.size() <= 0) {
                LOGGER.debug("We got the top ! (\"" + originalOuId + "\")");
                topLevelOuIds.add(originalOuId);
            } else if (parent.size() > 1) {
            	/**
            	 * TODO: rajouter le comptage sur l'identifiant du hiérarchique associé à la structure pour pouvoir déterminer
            	 * celui qui a le plus de subordonnés pour définir la structure d'attrachement N+1
            	 */ 
                LOGGER.warn("There is a big problem with the structure identified by \"" + originalOuId
                        + "\": more than one parent !");
            } else {
                String parentId = (String) parent.get(0);
                // Cas où sites.parent_id = sites.id pour identifier la structure de plus haut niveau 
                //if(parentId.compareToIgnoreCase(originialOuId) == 0) {
                //	LOGGER.debug("We got the top ! (\"" + originalOuId + "\")");
                //	topLevelOuId = originalOuId;
                //}
                if (ouChildren.containsKey(parentId)) {
                    ((List<String>) ouChildren.get(parentId)).add(originalOuId);
                } else {
                    List<String> values = new ArrayList<String>();
                    values.add(originalOuId);
                    ouChildren.put(parentId, values);
                }
            }
        }

        Iterator siIter = originalOuIdList.iterator();
        while (siIter.hasNext()) {
            String actualOuId = (String) siIter.next();
            fOrganizationalUnit st = _getOu(actualOuId);
            //if(topLevelOuIds.contains(st.getOu().toLowerCase())) {
            //if(topLevelOuIds.contains(actualOuId.toLowerCase())) {
            if(topLevelOuIds.contains(actualOuId)) {
                st.setDistinguishName("ou=" + st.getOu() + "," + getBaseDn());
            }
            ou.put(actualOuId, st);
        }

        initializeVector(topLevelOuIds);
    }

// SQL Methods

    /**
     * Do SQL request to get an organizationalUnit object.
     * @param ouId
     * @return An organizationalUnit object containing datas from database.
     */
    public fOrganizationalUnit _getOu(String ouId) {
		try {
			return (fOrganizationalUnit) sqlMapper.queryForObject(getSqlFunctionNameForGetObject(), ouId);
		} catch (SQLException e) {
            LOGGER.warn("Unable to found ou with id=" + ouId + " (" + e + ")", e);
		}
        return null;
    }

    /**
     * Do SQL request to get a list of members id from an ou specified by its id.
     * @return A list of members id.
     */
    public List _getOuMemberIdList(String ouId) {
        try {
        	List l = (List) sqlMapper.queryForList(getSqlFunctionNameForGetMembersList(), ouId) ;
        	this.lastOuMemberNb = l.size() ;
        	return l ;
		} catch (SQLException e) {
            LOGGER.warn("Unable to return ou member list (" + e + ")", e);
		}
		return null;
    }

    /**
     * Return ou member's number from last call of _getOuMemberIdList.
     * @return number of member's id.
     */
    public int _getNbOuMember() {
    	return this.lastOuMemberNb ;
    }

    /**
     * Do SQL request to get a list of organizationalUnit's id.
     * @return A list of organizationalUnit's id.
     */
    public List _getOuIdList() {
        try {
			return (List) sqlMapper.queryForList(getSqlFunctionNameForGetObjectsList(),null);
		} catch (SQLException e) {
            LOGGER.warn("Unable to return ou list (" + e + ")", e);
		}
		return null;
    }

    /**
     * Do SQL request to get an parent organizationalUnit id of LDAP entry specified by ouId.
     * @param ouId A LDAP organizationalUnit entry id.
     * @return A organizationalUnit id.
     */
    public List _getParent(String ouId) {
        try {
            return (List) sqlMapper.queryForList(getSqlFunctionNameForGetParentObjectId(), ouId);
		} catch (SQLException e) {
            LOGGER.warn("Unable to return ou parents list (" + e + ")", e);
		}
		return null;
    }

// Private Methods

    /**
     * Recursively add the items to the ordered vector
     * @param ouIds the object ids list
     */
    private void initializeVector(List<String> ouIds) {
    	Iterator<String> ouIdsIter = ouIds.iterator();
    	while(ouIdsIter.hasNext()) {
    		initializeVector(ouIdsIter.next());
    	}
    }
    
    /**
     * Recursively add the item to the ordered vector
     * @param ouId the object id
     */
    private void initializeVector(String ouId) {
        orderedOu.add(ouId);
        List children = (List)ouChildren.get(ouId);
        if(children != null) {
            Iterator childrenIter = children.iterator();
            while(childrenIter.hasNext()) {
                String childId = (String)childrenIter.next();
                fOrganizationalUnit childStructure = (fOrganizationalUnit) ou.get(childId);
                childStructure.setDistinguishName("ou=" + childStructure.getOu() + "," + ((fOrganizationalUnit)ou.get(ouId)).getDistinguishName());
                initializeVector(childId);
            }
        }
	}

	/**
	 * 
	 * @param id
	 * @return
	 */
	public fOrganizationalUnit getFlatObject(String id) {
		// Get it through the cache to have the dn setted by the cache initialize method
		return ou.get(id);
	}

	public String getSqlFunctionNameForGetParentObjectId() {
		return sqlFunctionNameForGetParentObjectId;
	}

	public String getSqlFunctionNameForGetMembersList() {
		return sqlFunctionNameForGetMembersList;
	}

	public String getSqlFunctionNameForGetObject() {
		return sqlFunctionNameForGetObject;
	}

	public String getSqlFunctionNameForGetObjectsList() {
		return sqlFunctionNameForGetObjectList;
	}

	public String getBaseDn() {
		return baseDn;
	}

}
